import requests
from . import exceptions as E


def call(method, url, **kwargs):
    """ Attachs the kwargs into the headers, sends the request to the Ford API's
        and handles all error cases.

    Args:
        method (str): HTTP method
        url (str): url of the request
        **kwargs: parameters for the request

    Returns:
        Response: response from the request to the Ford API's
    """

    if "headers" not in kwargs:
        kwargs["headers"] = {}
    kwargs['headers']['User-Agent'] = 'fordpass-na/353 CFNetwork/1121.2.2 Darwin/19.3.0'

    try:
        response = requests.request(method, url, timeout=310, **kwargs)
        code = response.status_code

        if response.ok:
            return response
        elif code == 400:
            raise E.ValidationException(response)
        elif code == 401:
            raise E.AuthenticationException(response)
        elif code == 403:
            raise E.PermissionException(response)
        elif code == 404:
            raise E.ResourceNotFoundException(response)
        elif code == 429:
            raise E.RateLimitingException(response)
        elif code == 500:
            raise E.ServerException(response)
        elif code == 504:
            raise E.GatewayTimeoutException(response)
        else:
            response.raise_for_status()

    except Exception as e:
        if isinstance(e, E.SyncException):
            raise e
        else:
            raise E.SyncException("Unexpected error") from e



API_URL = 'https://usapi.cv.ford.com/api'  # US Connected Vehicle api
VEHICLE_URL = 'https://services.cx.ford.com/api'
USER_URL = 'https://api.mps.ford.com/api'
TOKEN_URL = 'https://sso.ci.ford.com/oidc/endpoint/default/token'
APP_ID = '71A3AD0A-CF46-4CCF-B473-FC7FE5BC4592'



import requests


class SyncException(Exception):

    def __init__(self, response):
        self.message = 'Unknown error'
        if isinstance(response, requests.models.Response):
            json = response.json()
            if 'message' in json:
                self.message = json['message']
            elif 'error_description' in json:
                self.message = json['error_description']
        elif isinstance(response, str):
            self.message = response

    def __str__(self):
        return self.message


class ValidationException(SyncException):
    pass


class AuthenticationException(SyncException):
    pass


class PermissionException(SyncException):
    pass


class ResourceNotFoundException(SyncException):
    pass


class RateLimitingException(SyncException):
    pass


class ServerException(SyncException):
    pass


class GatewayTimeoutException(SyncException):
    def __init__(self, response):
        self.message = response.text

    def __str__(self):
        return self.message



"""
The Python package `connectedcar` provides a SDK to interact with
Ford Sync Connect enabled vehicles. With this package you will
be able to make requests on behalf of a Ford Pass user to: authorize
new vehicles, edit current vehicles, return vehicle information, send
vehicle commands and return user information.
"""

from .connectedcar import (AuthClient)
from .user import User
from .vehicle import Vehicle
from .exceptions import (
    SyncException,
    ValidationException,
    AuthenticationException,
    PermissionException,
    ResourceNotFoundException,
    RateLimitingException,
    ServerException,
    GatewayTimeoutException)



from .api import Api
from . import const


class Vehicle(object):

    def __init__(self, vehicle_id, access_token):
        """ Initialize a new Vehicle object to directly make requests to Ford.

        Args:
            vehicle_id (str): Vehicle identification number
            access_token (str): Sync Connect access token
        """
        self.access_token = access_token
        self.vehicle_id = vehicle_id
        self.api = Api(access_token)

    def status(self):
        """ GET Vehicle.status

        Returns:
            Response: Current vehicle status

        Raises:
            SyncException

        """

        response = self.api.get(
            const.API_URL,
            'vehicles/v4/' +
            self.vehicle_id +
            '/status?lrdt=01-01-1970%2000:00:00')
        return response.json()

    def refresh_status(self):
        """ PUT Vehicle.refresh_status

        Returns:
            Response: Send request to refresh data from the cars module

        Raises:
            SyncException

        """

        response = self.api.put(
            const.API_URL,
            'vehicles/v2/' +
            self.vehicle_id +
            '/status', None)
        return response.json()

    def send_auth(self):
        """ POST Vehicle.send_auth

        Returns:
            Response: Send vehicle authorization

        Raises:
            SyncException

        """

        response = self.api.post(
            const.API_URL, 'vehicles/v2/' + self.vehicle_id + '/drivers', None)
        return response.json()

    def auth_status(self):
        """ GET Vehicle.auth_status

        Returns:
            Response: Vehicle authorization status

        Raises:
            SyncException

        """

        response = self.api.get(
            const.API_URL,
            'vehicles/' +
            self.vehicle_id +
            '/authstatus?lrdt=01-01-1970%2000:00:00')
        return response.json()

    def details(self):
        """ GET Vehicle.detials

        Returns:
            Response: Vehicle details

        Raises:
            SyncException

        """

        response = self.api.get(
            const.API_URL,
            'users/vehicles/' +
            self.vehicle_id +
            '/detail?lrdt=01-01-1970%2000:00:00')
        return response.json()

    def maintenance_schedule(self):
        """ GET Vehicle.maintenance_schedule

        Returns:
            Response: Vehicle maintenance schedule

        Raises:
            SyncException

        """

        response = self.api.get(
            const.USER_URL,
            'vehiclemaintenance/v1/maintenance-schedule?vin=' +
            self.vehicle_id +
            '&language=EN&country=USA')
        return response.json()

    def capability(self):
        """ GET Vehicle.capability

        Returns:
            Response: Vehicle capabilities

        Raises:
            SyncException

        """

        response = self.api.get(
            const.USER_URL, 'capability/v1/vehicles/' + self.vehicle_id)
        return response.json()

    def vin(self):
        """ GET Vehicle.vin

        Returns:
            Response: Vehicle vin

        Raises:
            SyncException

        """

        json = self.status()
        return {"data": json['vehiclestatus']['vin']}

    def odometer(self):
        """ GET Vehicle.odometer

        Returns:
            Response: Vehicle odometer

        Raises:
            SyncException

        """

        json = self.status()
        return {"data": json['vehiclestatus']['odometer']}

    def fuel(self):
        """ GET Vehicle.fuel

        Returns:
            Response: Vehicle fuel level

        Raises:
            SyncException

        """

        json = self.status()
        return {"data": json['vehiclestatus']['fuel']}

    def oil(self):
        """ GET Vehicle.oil

        Returns:
            Response: Vehicle oil life

        Raises:
            SyncException

        """

        json = self.status()
        return {"data": json['vehiclestatus']['oil']}

    def tire_pressure(self):
        """ GET Vehicle.tire_pressure

        Returns:
            Response: Vehicle tire pressure

        Raises:
            SyncException

        """

        json = self.status()
        return {"data": json['vehiclestatus']['TPMS']}

    def battery(self):
        """ GET Vehicle.battery

        Returns:
            Response: Vehicle battery status

        Raises:
            SyncException

        """

        json = self.status()
        return {"data": json['vehiclestatus']['battery']}

    def location(self):
        """ GET Vehicle.location

        Returns:
            Response: Vehicle location

        Raises:
            SyncException

        """

        json = self.status()
        return {"data": json['vehiclestatus']['gps']}

    def window_positions(self):
        """ GET Vehicle.window_position

        Returns:
            Response: Vehicle window positions

        Raises:
            SyncException

        """

        json = self.status()
        return {"data": json['vehiclestatus']['windowPosition']}

    def door_status(self):
        """ GET Vehicle.door_status

        Returns:
            Response: Vehicle doors status

        Raises:
            SyncException

        """

        json = self.status()
        return {"data": json['vehiclestatus']['doorStatus']}

    def lock_status(self):
        """ GET Vehicle.lock_Status

        Returns:
            Response: Vehicle lock status

        Raises:
            SyncException

        """

        json = self.status()
        return {"data": json['vehiclestatus']['lockStatus']}

    def alarm_status(self):
        """ GET Vehicle.alarm_status

        Returns:
            Response: Vehicle alarm status

        Raises:
            SyncException

        """

        json = self.status()
        return {"data": json['vehiclestatus']['alarm']}

    def ignition_status(self):
        """ GET Vehicle.ignition_status

        Returns:
            Response: Vehicle ignition status

        Raises:
            SyncException

        """

        json = self.status()
        return {"data": json['vehiclestatus']['ignitionStatus']}

    def wakeup(self):
        """ GET Vehicle.wakeup

        Returns:
            Response: response from the request to the Ford API

        Raises:
            SyncException

        """

        response = self.api.action(
            'GET',
            const.USER_URL,
            'dashboard/v1/users/vehicles?language=EN&wakeupVin=' +
            self.vehicle_id +
            '&skipRecall=true&country=USA&region=US')
        return response.json()

    def start(self):
        """ PUT/GET Vehicle.start

        Returns:
            Response: response from the request to the Ford API

        Raises:
            SyncException

        """

        job_id = self.api.action(
            'PUT',
            const.API_URL,
            'vehicles/v2/' +
            self.vehicle_id +
            '/engine/start').json()['commandId']
        response = self.api.action(
            'GET',
            const.API_URL,
            'vehicles/' +
            self.vehicle_id +
            '/engine/start/' +
            job_id)
        return response.json()

    def stop(self):
        """ DELETE/GET Vehicle.stop

        Returns:
            Response: response from the request to the Ford API

        Raises:
            SyncException

        """

        job_id = self.api.action(
            'DELETE',
            const.API_URL,
            'vehicles/v2/' +
            self.vehicle_id +
            '/engine/start').json()['commandId']
        response = self.api.action(
            'GET',
            const.API_URL,
            'vehicles/' +
            self.vehicle_id +
            '/engine/start/' +
            job_id)
        return response.json()

    def lock(self):
        """ PUT/GET Vehicle.lock

        Returns:
            Response: response from the request to the Ford API

        Raises:
            SyncException

        """

        job_id = self.api.action(
            'PUT',
            const.API_URL,
            'vehicles/v2/' +
            self.vehicle_id +
            '/doors/lock').json()['commandId']
        response = self.api.action(
            'GET',
            const.API_URL,
            'vehicles/' +
            self.vehicle_id +
            '/doors/lock/' +
            job_id)
        return response.json()

    def unlock(self):
        """ DELETE/GET Vehicle.unlock

        Returns:
            Response: response from the request to the Ford API

        Raises:
            SyncException

        """

        job_id = self.api.action(
            'DELETE',
            const.API_URL,
            'vehicles/v2/' +
            self.vehicle_id +
            '/doors/lock').json()['commandId']
        response = self.api.action(
            'GET',
            const.API_URL,
            'vehicles/' +
            self.vehicle_id +
            '/doors/lock/' +
            job_id)
        return response.json()



from .api import Api
from . import const


class User(object):

    def __init__(self, access_token):
        """ Initialize a new User object to directly make requests to Ford.

        Args:
            access_token (str): Ford access token
        """
        self.api = Api(access_token)

    def info(self):
        """ GET User.info

        Returns:
            Response: User information

        """

        response = self.api.get(const.API_URL, 'users')
        return response.json()

    def get_messages(self):
        """ GET User.messages

        Returns:
            Response: User messages from Ford API

        """

        response = self.api.get(
            const.USER_URL, 'messagecenter/v3/messages?lrdt=1970-01-01T00:00:00Z')
        return response.json()

    def vehicles(self):
        """ GET User.vehicles

        Returns:
            Response: User vehicle list

        """

        response = self.api.get(
            const.VEHICLE_URL,
            'dashboard/v1/users/vehicles?country=USA&language=EN&region=US&skipRecall=true')
        return response.json()

    def add_vehicle(self, vehicle_id):
        """ POST User.add_vehicle

        Args:
            vehicle_id (str): the vehicle identification number

        Returns:
            Response: response from the request to the Ford API

        """

        data = '{"countryCode": "USA", "nickName": "", "vin": "' + \
            vehicle_id + '", "appBrand": "F", "appRegion": "NA"}'
        response = self.api.post(const.USER_URL, 'garage/mobile', data)
        return response.json()

    def delete_vehicle(self, vehicle_id):
        """ DELETE User.delete_vehicle

        Args:
            vehicle_id (str): the vehicle identification number

        Returns:
            Response: response from the request to the Ford API

        """

        response = self.api.delete(
            const.API_URL, 'users/vehicles/' + vehicle_id)
        return response.json()

    def request_vehicle_access(self, vehicle_id):
        """ POST User.accessrequest

        Args:
            vehicle_id (str): the vehicle identification number

        Returns:
            Response: Request access to user vehicle with vehicle_id

        """

        response = self.api.post(
            const.API_URL, 'vehicles/${vin}/drivers/accessrequest', {})
        return response.json()

    def authorize_vehicle_access(self, message_id):
        """ PUT User.accessrequest

        Args:
            vehicle_id (str): the vehicle identification number

        Returns:
            Response: Request access to user vehicle with vehicle_id

        """

        response = self.api.put(
            const.USER_URL, 'authorization/retail/v1/' + message_id + '/approveAccess', {})
        return response.json()



from . import const, requester


class Api(object):

    def __init__(self, access_token):
        """ Initialize a new Api object to directly make requests to Ford.

        Args:
            access_token (str): Ford access token
        """
        self.access_token = access_token
        self.auth = {
            'auth-token': access_token,
            'Accept': '*/*',
            'Accept-Language': 'en-US',
            'User-Agent': 'fordpass-na/353 CFNetwork/1121.2.2 Darwin/19.3.0',
            'Content-Type': 'application/json',
            'Accept-Encoding': 'gzip, deflate, br',
            'Application-Id': const.APP_ID
        }

    def _format(self, endpoint, context):
        """ Generates the formated URL

        Args:
            endpoint (str): the Ford endpoint of interest
            context (str): the API endpoint context

        Returns:
            str: formatted url
        """

        return '{}/{}'.format(endpoint, context)

    def get(self, endpoint, context):
        """ Sends GET requests to Ford

        Args:
            endpoint (str): the Ford endpoint of interest
            context (str): the API endpoint context

        Returns:
            Response: response from the request to the Ford API
        """

        url = self._format(endpoint, context)
        headers = self.auth

        return requester.call('GET', url, headers=headers)

    def post(self, endpoint, context, data):
        """ Sends POST requests to Ford

        Args:
            endpoint (str): the Ford endpoint of interest
            context (str): the API endpoint context
            data (array): data to be sent with request

        Returns:
            Response: response from the request to the Ford API
        """

        url = self._format(endpoint, context)
        headers = self.auth

        return requester.call('POST', url, headers=headers, data=data)

    def put(self, endpoint, context, data):
        """ Sends PUT requests to Ford

        Args:
            endpoint (str): the Ford endpoint of interest
            context (str): the API endpoint context
            data (array): data to be sent with request

        Returns:
            Response: response from the request to the Ford API
        """

        url = self._format(endpoint, context)
        headers = self.auth

        return requester.call('PUT', url, headers=headers, data=data)

    def delete(self, endpoint, context):
        """ Sends DELETE requests to Ford

        Args:
            endpoint (str): the Ford endpoint of interest
            context (str): the API endpoint context

        Returns:
            Response: response from the request to the Ford API
        """

        url = self._format(endpoint, context)
        headers = self.auth

        return requester.call('DELETE', url, headers=headers)

    def action(self, method, endpoint, context):
        """ Sends universal requests to Ford

        Args:
            method (str): the request type
            endpoint (str): the Ford endpoint of interest
            context (str): the API endpoint context

        Returns:
            Response: response from the request to the Ford API
        """

        url = self._format(endpoint, context)
        headers = self.auth

        return requester.call(method, url, headers=headers)



from . import const, requester


class AuthClient(object):

    def __init__(self, client_id, client_secret,
                 redirect_uri=None, scope=None):
        """ A client for accessing the Ford API

        Args:
            client_id (str): The application id, provided in the application
                dashboard
            client_secret (str): The application secret, provided in the
                application dashboard
            redirect_uri (str, optional): The URL to redirect to after the user accepts
                or declines the application's permissions.
            scope (list, optional): A list of permissions requested by the application

        """
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri

    def get_user_access_token(self, username, password):
        """ Exchange a username and password for a new access dictionary

        Args:
            username (str): Ford pass username
            password (str): Ford pass password

        Returns:
            Response: response containing access and refresh token

        Raises:
            SyncException
        """

        headers = {
            'Accept': '*/*',
            'Accept-Language': 'en-US',
            'Content-Type': 'application/x-www-form-urlencoded',
            'User-Agent': 'fordpass-na/353 CFNetwork/1121.2.2 Darwin/19.3.0',
            'Accept-Encoding': 'gzip, deflate, br'
        }

        data = {
            'client_id': self.client_id,
            'grant_type': 'password',
            'username': username,
            'password': password
        }

        response = requester.call(
            'POST', const.TOKEN_URL, headers=headers, data=data).json()
        return response

    def exchange_refresh_token(self, refresh_token):
        """ Exchange a refresh token for a new access dictionary

        Args:
            refresh_token (str): A valid refresh token from a previously retrieved
                access object

        Returns:
            Response: response containing access and refresh token

        Raises:
            SyncException
        """

        headers = {
            'Accept': '*/*',
            'Accept-Language': 'en-US',
            'Content-Type': 'application/x-www-form-urlencoded',
            'User-Agent': 'fordpass-na/353 CFNetwork/1121.2.2 Darwin/19.3.0',
            'Accept-Encoding': 'gzip, deflate, br'
        }

        data = {
            'client_id': self.client_id,
            'grant_type': 'refresh_token',
            'refresh_token': refresh_token
        }

        response = requester.call(
            'POST', const.TOKEN_URL, headers=headers, data=data).json()
        return response



from os.path import dirname, isdir, join
from setuptools import setup
import subprocess
import os
import re


def _get_version():
    """Extract version from git."""
    version_re = re.compile('^Version: (.+)$', re.M)
    d = dirname(__file__)

    if isdir(join(d, '.git')):
        # Get the version using "git describe".
        cmd = 'git describe --tags --match [0-9]*'.split()

        try:
            version = subprocess.check_output(cmd).decode().strip()
        except subprocess.CalledProcessError:
            print('Unable to get version number from git tags')
            exit(1)

        if '-' in version:
            version = '.post'.join(version.split('-')[:2])

        with open(os.devnull, 'w') as fd_devnull:
            subprocess.call(['git', 'status'],
                            stdout=fd_devnull, stderr=fd_devnull)

        cmd = 'git diff-index --name-only HEAD'.split()

        try:
            dirty = subprocess.check_output(cmd).decode().strip()
        except subprocess.CalledProcessError:
            print('Unable to get git index status')
            exit(1)

        if dirty != '':
            version += '.dev1'

    else:
        # Extract the version from the PKG-INFO file.
        with open(join(d, 'PKG-INFO')) as f:
            version = version_re.search(f.read()).group(1)

    return version


def _get_long_description():
    """Get README contents."""
    with open('README.md') as reader:
        return reader.read()


setup(
    name='connectedcar',
    version=_get_version(),
    description='ConnectedCar SDK',
    long_description=_get_long_description(),
    long_description_content_type='text/markdown',
    author='Ian White',
    author_email='iwhite99@protonmail.com',
    packages=['connectedcar'],
    url='https://github.com/ianjwhite99/connected-car-python-sdk',
    license='MIT',
    install_requires=[
        'requests'
    ],
    extras_require={
        "dev": [
            "coverage",
            "nose",
            "responses",
            "pylint",
            "python-dotenv",
            "requests"
        ]
    },
)
